from __future__ import annotations

import abc
import enum
import os.path
import time
import subprocess
from .. import Devutils as dev
from . import ManagedJobQueues as queues

__all__ = [
    "ExecutionStatus",
    "ExecutionQueue",
    "ExecutionEngine",
    # "PoolSubmissionEngine",
    "ManagedJobQueueExecutionEngine",
    "SLURMExecutionEngine"
]

class ExecutionStatus(enum.Enum):
    UNKNOWN = "unknown"
    RUNNING = "running"
    PENDING = "pending"
    COMPLETED = "completed"
    ERROR = "error"

class ExecutionFuture(metaclass=abc.ABCMeta):
    poll_time = 1
    def __init__(self, poll_time=None):
        self.status = ExecutionStatus.UNKNOWN
        self._is_complete = False
        self._result = None
        self.poll_time = self.poll_time if poll_time is None else poll_time
    def join(self, timeout=None):
        self.status = self.get_status()
        cur = time.time()
        while self.status in {ExecutionStatus.UNKNOWN, ExecutionStatus.RUNNING, ExecutionStatus.PENDING}:
            time.sleep(self.poll_time)
            self.status = self.get_status()
            if timeout is not None:
                now = time.time()
                if now - cur > timeout:
                    raise TimeoutError(f"job {self} timed out during join")
    @abc.abstractmethod
    def get_result(self) -> dict:
        ...
    @abc.abstractmethod
    def get_status(self) -> ExecutionStatus:
        ...

class JoinableExecutionFuture(ExecutionFuture):
    @abc.abstractmethod
    def await_result(self, timeout=None):
        ...
    def join(self, timeout=None):
        self.await_result(timeout=timeout)

class ExecutionQueue:
    def __init__(self, futures:list[ExecutionFuture]):
        self.futures = futures
    def join(self, timeout=None):
        for j in self.futures:
            cur = time.time()
            j.join(timeout=timeout)
            if timeout is not None:
                timeout = timeout - (time.time() - cur)
                if timeout < 0:
                    raise TimeoutError(f"pool {self} timed out during join")

class ExecutionEngine(metaclass=abc.ABCMeta):
    name: str
    future_type: type[ExecutionFuture]
    engine_types = {}
    @classmethod
    def register(cls, name, engine=None):
        if engine is None:
            if isinstance(name, str):
                def register(engine):
                    return cls.register(name, engine)
                return register
            elif isinstance(name, ExecutionEngine):
                engine = name
                return cls.register(engine.name, engine)
        else:
            cls.engine_types[name] = engine
        return engine
    @classmethod
    def resolve(cls, name, **opts):
        return cls.engine_types[name](**opts)

    def __init__(self, **opts):
        self._call_depth = 0
        self.opts = opts

    @abc.abstractmethod
    def submit_job(self, **kwargs) -> ExecutionFuture:
        ...

    def submit_jobs(self, jobs: list[dict], **kwargs) -> ExecutionQueue:
        return ExecutionQueue([
            self.submit_job(**dict(job, **kwargs))
            for job in jobs
        ])

    def __enter__(self):
        # here to be extended
        self._call_depth += 1
        if self._call_depth == 1:
            self.startup()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._call_depth -= 1
        if self._call_depth < 1:
            self.shutdown()

    def startup(self):
        ...

    def shutdown(self):
        ...

class FileBackedExecutionFuture(ExecutionFuture):
    results_file = 'results.json'
    status_file = 'status.json'
    def __init__(self,
                 watch_dir=None,
                 poll_time=None,
                 results_file=None,
                 status_file=None):
        super().__init__(poll_time=poll_time)
        self.watch_dir = watch_dir
        if results_file is None:
            results_file = self.results_file
        self.results_file = results_file
        if status_file is None:
            status_file = self.status_file
        self.status_file = status_file
    def get_result(self) -> dict:
        if self.watch_dir is None:
            results_file = self.results_file
        else:
            results_file = os.path.join(self.watch_dir, self.results_file)
        return dev.read_json(results_file)
    def get_status(self) -> ExecutionStatus:
        if self.watch_dir is None:
            status_file = self.status_file
        else:
            status_file = os.path.join(self.watch_dir, self.status_file)
        if os.path.isfile(status_file):
            stat_dict = dev.read_json(status_file)
            return ExecutionStatus(stat_dict.get('status', 'UNKNOWN'))
        else:
            return ExecutionStatus.UNKNOWN

class ManagedJobQueueExecutionFuture(FileBackedExecutionFuture):
    def __init__(self,
                 job_id,
                 queue_manager:queues.ManagedJobQueueHandler,
                 watch_dir=None,
                 results_file=None,
                 status_file=None,
                 poll_time=None):
        super().__init__(
            watch_dir=watch_dir, poll_time=poll_time,
            results_file=results_file,
            status_file=status_file
        )
        self.queue_manager = queue_manager
        self.job_id = job_id

    queue_status_map = {
        queues.ManagedJobQueueJobStatus.PENDING:ExecutionStatus.PENDING,
        queues.ManagedJobQueueJobStatus.COMPLETED:ExecutionStatus.COMPLETED,
        queues.ManagedJobQueueJobStatus.RUNNING:ExecutionStatus.RUNNING,
        queues.ManagedJobQueueJobStatus.ERROR:ExecutionStatus.ERROR,
    }
    def get_status(self) -> ExecutionStatus:
        return self.queue_status_map[
            self.queue_manager.get_job_status(self.job_id)
        ]

class ManagedJobQueueExecutionEngine(ExecutionEngine):
    future_type = ManagedJobQueueExecutionFuture
    def __init__(self, queue_manager:queues.ManagedJobQueueHandler, **opts):
        super().__init__(**opts)
        self.queue_manager = queue_manager

    def prep_future_opts(self,
                         watch_dir=None,
                         results_file=None,
                         status_file=None,
                         poll_time=None,
                         **kwargs):
        return dict(
            watch_dir=watch_dir,
            results_file=results_file,
            status_file=status_file,
            poll_time=poll_time
        ), kwargs

    def submit_job(self, *,
                   watch_dir=None, poll_time=None,
                   results_file=None,
                   status_file=None,
                   **kwargs) -> ManagedJobQueueExecutionFuture:
        #TODO: track output files
        fut_opts, job_opts = self.prep_future_opts(**kwargs)
        id = self.queue_manager.submit_job(**job_opts)
        return self.future_type(
            id,
            self.queue_manager,
            **fut_opts
        )

class SLURMExecutionFuture(ManagedJobQueueExecutionFuture):
    def get_status(self) -> ExecutionStatus:
        if self.status in {ExecutionStatus.RUNNING, ExecutionStatus.COMPLETED, ExecutionStatus.ERROR}:
            try:
                return super().get_status()
            except KeyError: # job no longer exists because it's done or errored
                return ExecutionStatus.COMPLETED
        else:
            return super().get_status()

class SLURMExecutionEngine(ManagedJobQueueExecutionEngine):
    future_type = SLURMExecutionFuture
    def __init__(self, **opts):
        super().__init__(queues.SLURMHandler(), **opts)

    def prep_future_opts(self, *,
                         sbatch_file,
                         watch_dir=None,
                         chdir=None, **kwargs):
        if watch_dir is None:
            if chdir is None:
                watch_dir = os.path.dirname(sbatch_file)
            else:
                watch_dir = chdir

        return super().prep_future_opts(
            watch_dir=watch_dir,
            chdir=chdir,
            sbatch_file=sbatch_file,
            **kwargs
        )

    def submit_job(self,
                   sbatch_file,
                   *,
                   watch_dir=None, poll_time=None,
                   results_file=None,
                   status_file=None,
                   **kwargs) -> ManagedJobQueueExecutionFuture:
        #TODO: track output files
        fut_opts, job_opts = self.prep_future_opts(sbatch_file=sbatch_file, **kwargs)
        id, _ = self.queue_manager.submit_job(**job_opts)
        return self.future_type(
            id,
            self.queue_manager,
            **fut_opts
        )

class ProcessExecutionFuture(JoinableExecutionFuture):
    def __init__(self, base_obj, **ignored):
        super().__init__(**ignored)
        self.obj = base_obj
        self._get_result = dev.default
    def await_result(self, timeout=None):
        if dev.is_default(self._get_result, allow_None=False):
            self._get_result = self.obj.get(timeout=timeout)
    def get_result(self):
        if dev.is_default(self._get_result):
            return None
        else:
            return self._get_result
    def get_status(self) -> ExecutionStatus:
        try:
            res = self.obj.successful()
        except (ValueError, AssertionError):
            return ExecutionStatus.RUNNING
        else:
            if res:
                return ExecutionStatus.COMPLETED
            else:
                return ExecutionStatus.ERROR

class ProcessGeneratorExecutionEngine(ExecutionEngine):
    future_type = ProcessExecutionFuture
    def __init__(self, proc_gen, **opts):
        super().__init__(**opts)
        self.proc_gen = proc_gen

    def submit_job(self, method, **kwargs):
        # TODO: track output files
        proc = self.proc_gen.apply_async(method, **kwargs)
        return self.future_type(
            proc
        )

    def startup(self):
        self.proc_gen.__enter__()

    def shutdown(self):
        self.proc_gen.__exit__()


import uuid

from .JHTML import HTML

__all__ = [
    "JSMol"
]

__reload_hooks__ = [".JHTML"]

class JSMol:
    class Applet(HTML.Div):
        version = "16.3.7.9"
        jsmol_source = f"https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@{version}/jsmol/JSmol.min.js"
        jmol2_source = f"https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@{version}/jsmol/js/Jmol2.js"
        patch_script = f"""
        if (typeof Jmol._patched === 'undefined') {{
            Jmol._patched = true;
            Jmol._serverUrl = Jmol.Info["serverURL"];
            Jmol._appletNameMap = {{}};
            jmolInitialize('https://cdn.jsdelivr.net/gh/b3m2a1/jsmol-cdn@{version}/jsmol/');
        }};
        """
        unsynced_properties = ['width', 'height']
        # can_by_dynamic = False
        @classmethod
        def load_applet_script(cls, id, loader,
                               include_script_interface=False,
                               interface_target="",
                               target=None):
            if target is None:
                target = id

            injection = "''"
            if include_script_interface:
                from ..Plots.X3DInterface import X3D

                input_tag = id + "-script-input"

                elems = []
                console = HTML.Textarea(id=input_tag, width='100%')
                elems.append(console)
                button = HTML.Button("Run Script", id=id + "-button-input",
                                onclick=f"""
                (function() {{
                    const script = document.getElementById('{input_tag}').value;
                    const app = Jmol._appletNameMap['{id}'];
                    app._script(script);
                    document.getElementById('{input_tag}').value = '';
                }})()
                """
                                )

                new_id = 'jmolApplet_' + id.split("-")[-2]
                elems.append(button)
                elems.append(
                    HTML.Div([
                        HTML.Button("Save Figure", onclick=X3D.get_export_script(new_id + '_appletdiv')),
                        HTML.Button("Record Animation", onclick=X3D.get_record_screen_script(new_id+ '_appletdiv')),
                        HTML.Input(value="2", id=id + '-duration-input',
                                   oninput=X3D.set_animation_duration_script(id))
                    ], style='display:flex')
                )

                strings = "\n\n".join([e.tostring() for e in elems])

                injection = f"`<div style='display:block'>\n{strings}\n</div>`"
            load_script = f"""
(function() {{
   $.getScript('{cls.jmol2_source}').then(
   () => {{
       {cls.patch_script}
       let loaded = false;
        if (!loaded) {{
            if (typeof Jmol._appletNameMap === "undefined") {{
                Jmol._appletNameMap = {{}}
            }};
            loaded = true;
            let applet = {loader};
            applet.serverURL = Jmol.Info.serverURL;
            let wrapper = document.getElementById('{id}');
            wrapper.innerHTML = applet._code;
            Jmol._appletNameMap['{id}'] = applet;
            wrapper.ondelete = function() {{ delete Jmol._appletNameMap['{id}'] }};
            if ('{interface_target}'.length > 0) {{
                document.getElementById('{interface_target}').innerHTML = {injection};
            }}
        }}
    }})
}})();
"""
            base_script = HTML.Script(src=cls.jsmol_source,
                               onload=load_script
                               )
            return base_script

        def __init__(self, *model_etc, width=500, height=500,
                     animate=False, vibrate=False,
                     load_script=None,
                     suffix=None,
                     id=None,
                     dynamic_loading=None,
                     include_script_interface=False,
                     create_applet_loader=None,
                     style=None,
                     **attrs):
            if suffix is None:
                suffix = str(uuid.uuid4())[:6].replace("-", "")
            self.suffix = suffix
            if id is None:
                id =  "jsmol-applet-" + self.suffix
            self.id = id
            if len(model_etc) > 0:
                if isinstance(model_etc[0], str):
                    model_file = model_etc[0]
                    rest = model_etc[1:]
                    if create_applet_loader is None:
                        create_applet_loader = True
                else:
                    model_file = None
                    if create_applet_loader is None:
                        create_applet_loader = False
                    rest = model_etc
            else:
                model_file = None
                if create_applet_loader is None:
                    create_applet_loader = False
                rest = model_etc
            if len(rest) == 1 and isinstance(rest[0], (list, tuple)):
                rest = rest[0]

            if load_script is None:
                load_script = []
            if isinstance(load_script, str):
                load_script = [load_script]
            load_script = list(load_script)

            if animate:
                load_script.extend(["anim mode palindrome", "anim on"])
            elif vibrate:
                load_script.append("vibration on")

            if dynamic_loading is None:
                from ..Jupyter.JHTML import JupyterAPIs
                dynamic_loading = JupyterAPIs().in_jupyter_environment()
            self.dynamic_loading = dynamic_loading

            self.load_script = load_script
            self.width, self.height = width, height
            if create_applet_loader:
                elems = self.create_applet(model_file, include_script_interface=include_script_interface) + list(rest)
            else:
                elems = rest
            if include_script_interface:
                height = height + 200
            if style is not None:
                if 'width' not in style:
                    style['width'] =  f'{width}px'
                if 'height' not in style:
                    style['height'] =  f'{height}px'
            else:
                attrs['width'] =  f'{width}px'
                attrs['height'] = f'{height}px'
            super().__init__(*elems, id=self.id, style=style, **attrs)

        @property
        def applet_target(self):
            return f"_{self.suffix}"
        def prep_load_script(self):
            return '; '.join(self.load_script)
        def create_applet(self, model_file, include_script_interface=False):
            targ = self.applet_target
            width, height = self.width, self.height
            load_script = self.prep_load_script()
            if model_file is None:
                loader = f"jmolApplet([{width}, {height}], 'load {model_file}; {load_script}', '{targ}')"
            elif (
                    model_file.startswith("https://")
                    or model_file.startswith("file://")
                    or model_file.startswith("http://")
            ):
                loader = f"jmolApplet([{width}, {height}], 'load {model_file}; {load_script}', '{targ}')"
            else:
                loader = f"jmolAppletInline([{width}, {height}], `{model_file}`, '{load_script}', '{targ}')"

            kill_id = "tmp-" + str(uuid.uuid4())[:10]
            if include_script_interface:
                replacement_target = self.id+'-applet'
                interface_target = self.id+'-interface'
            else:
                replacement_target = self.id
                interface_target = ""
            load_script = self.load_applet_script(replacement_target,
                                                  loader,
                                                  target=targ,
                                                  interface_target=interface_target,
                                                  include_script_interface=include_script_interface)

            if self.dynamic_loading:
                load_script = load_script.tostring().replace("`", "\`")
                loader = HTML.Image(
                        src='data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7',
                        id=kill_id,
                        onload=f"""
                            (function() {{
                                let killElem = document.getElementById("{kill_id}");
                                if (killElem !== null) {{
                                    killElem.remove();
                                    const frag = document.createRange().createContextualFragment(`{load_script}`);
                                    document.head.appendChild(frag);
                                }}
                            }})()"""
                    )
            else:
                loader = load_script

            if include_script_interface:
                loader = [
                    HTML.Div(loader, width='100%', height=f'{self.height}px', id=replacement_target),
                    HTML.Div(height='200px', width='100%', padding='2rem', id=interface_target)
                ]
            else:
                loader = [loader]

            return loader

        def show(self):
            return self.display()



import numpy as np
from collections import namedtuple
from .Parsers import ElectronicStructureLogReader
from ...Parsers import FileStreamReader, Number, RegexPattern, Integer, Word

__all__ = [
    "MOLPROLogReader"
]


class MOLPROLogReader(ElectronicStructureLogReader):
    components_name = "MOLPROLogComponents"